# Unit test predict ForecasterEquivalentDate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from skforecast.recursive import ForecasterEquivalentDate

# Fixtures
from .fixtures_forecaster_equivalent_date import y


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterEquivalentDate(offset=1, n_offsets=1)

    err_msg = re.escape(
                ('This Forecaster instance is not fitted yet. Call `fit` with '
                 'appropriate arguments before using predict.')
              )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict(steps=3)


def test_predict_5_steps_with_int_offset_offset_1_n_offsets_1():
    """
    Test predict method with int offset, offset=1 and n_offsets=1.
    """
    forecaster = ForecasterEquivalentDate(
                     offset    = 1,
                     n_offsets = 1
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=5)

    expected = pd.Series(
        data  = np.array([0.61289453, 0.61289453, 0.61289453, 0.61289453, 0.61289453]),
        index = pd.date_range(start='2000-02-20', periods=5, freq='D'),
        name  = 'pred'
    )

    pd.testing.assert_series_equal(predictions, expected)


def test_predict_5_steps_with_int_offset_7_n_offsets_1():
    """
    Test predict method with int offset, offset=7 and n_offsets=1.
    """
    forecaster = ForecasterEquivalentDate(
                     offset    = 7,
                     n_offsets = 1
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=5)

    expected = pd.Series(
        data  = np.array([0.41482621, 0.86630916, 0.25045537, 0.48303426, 0.98555979]),
        index = pd.date_range(start='2000-02-20', periods=5, freq='D'),
        name  = 'pred'
    )

    pd.testing.assert_series_equal(predictions, expected)


def test_predict_7_steps_with_int_offset_7_n_offsets_2():
    """
    Test predict method with int offset, offset=7 and n_offsets=2.
    """
    forecaster = ForecasterEquivalentDate(
                     offset    = 7,
                     n_offsets = 2,
                     agg_func  = np.mean
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=7)

    expected = pd.Series(
        data  = np.array([0.42058876, 0.87984916, 0.5973077 , 0.49243547, 0.80475637,
                          0.31755176, 0.46509001]),
        index = pd.date_range(start='2000-02-20', periods=7, freq='D'),
        name  = 'pred'
    )

    pd.testing.assert_series_equal(predictions, expected)


def test_predict_7_steps_with_int_offset_5_n_offsets_1_using_last_window():
    """
    Test predict method with int offset, offset=1 and n_offsets=1, using last_window.
    """
    last_window = pd.Series(
        data  = np.array([0.73799541, 0.18249173, 0.17545176, 0.53155137, 0.53182759]),
        index = pd.date_range(start='2000-08-10', periods=5, freq='D'),
        name  = 'y'
    )

    forecaster = ForecasterEquivalentDate(
                     offset    = 5,
                     n_offsets = 1
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=7, last_window=last_window)

    expected = pd.Series(
        data  = np.array([0.73799541, 0.18249173, 0.17545176, 0.53155137, 0.53182759,
                          0.73799541, 0.18249173]),
        index = pd.date_range(start='2000-08-15', periods=7, freq='D'),
        name  = 'pred'
    )

    pd.testing.assert_series_equal(predictions, expected)


def test_predict_7_steps_with_int_offset_5_n_offsets_2_using_last_window():
    """
    Test predict method with int offset, offset=1 and n_offsets=1, using last_window.
    """
    last_window = pd.Series(
        data = np.array([0.73799541, 0.18249173, 0.17545176, 0.53155137, 0.53182759,
                         0.25045537, 0.48303426, 0.98555979, 0.51948512, 0.61289453]),
        index = pd.date_range(start='2000-08-05', periods=10, freq='D'),
        name = 'y'
    )
    forecaster = ForecasterEquivalentDate(
                     offset    = 5,
                     n_offsets = 2,
                     agg_func  = np.mean
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=7, last_window=last_window)
    
    expected = pd.Series(
        data  = np.array([0.49422539, 0.332763, 0.58050578, 0.52551824, 0.57236106,
                          0.49422539, 0.332763]),
        index = pd.date_range(start='2000-08-15', periods=7, freq='D'),
        name  = 'pred'
    )

    pd.testing.assert_series_equal(predictions, expected)


def test_predict_ValueError_when_all_equivalent_values_are_missing():
    """
    Test ValueError when equivalent values are missing because the offset is
    too large.
    """
    y = pd.Series(np.arange(7), 
                  index=pd.date_range(start='2000-02-25', periods=7, freq='D'))

    forecaster = ForecasterEquivalentDate(
                     offset    = pd.offsets.MonthEnd(1),
                     n_offsets = 1
                 )
    forecaster.fit(y=y)

    last_window = pd.Series(np.arange(5), 
                            index=pd.date_range(start='2000-02-10', periods=5, freq='D'))

    err_msg = re.escape(
                (f"All equivalent values are missing. This is caused by using "
                 f"an offset (<MonthEnd>) larger than the available data. "
                 f"Try to decrease the size of the offset (<MonthEnd>), "
                 f"the number of n_offsets (1) or increase the "
                 f"size of `last_window`. In backtesting, this error may be "
                 f"caused by using an `initial_train_size` too small.")
              )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict(steps=3, last_window=last_window)


def test_predict_ValueError_when_any_equivalent_values_are_missing():
    """
    Test ValueError when equivalent values are missing because the offset is
    too large.
    """
    y = pd.Series(np.arange(51), 
                  index=pd.date_range(start='2000-02-25', periods=51, freq='D'))
    
    forecaster = ForecasterEquivalentDate(
                     offset    = pd.offsets.MonthEnd(1),
                     n_offsets = 2,
                     agg_func  = np.mean
                 )
    forecaster.fit(y=y)

    last_window = pd.Series(np.arange(50), 
                            index=pd.date_range(start='2000-02-10', periods=50, freq='D'))

    warn_msg = re.escape(
                (f"Steps: ['2000-03-31', '2000-04-01', '2000-04-02'] "
                 f"are calculated with less than 2 n_offsets. "
                 f"To avoid this, increase the `last_window` size or decrease "
                 f"the number of n_offsets. The current configuration requires " 
                 f"a total offset of <2 * MonthEnds>.")
              )
    with pytest.warns(UserWarning, match = warn_msg):
        forecaster.predict(steps=3, last_window=last_window)


def test_predict_7_steps_with_offset_Day_5_n_offsets_1_using_last_window():
    """
    Test predict method with offset pd.offsets.Day() and 
    n_offsets=1, using last_window.
    """
    last_window = pd.Series(
        data = np.array([0.73799541, 0.18249173, 0.17545176, 0.53155137, 0.53182759]),
        index = pd.date_range(start='2000-08-10', periods=5, freq='D'),
        name = 'y'
    )

    forecaster = ForecasterEquivalentDate(
                     offset    = pd.offsets.Day(5),
                     n_offsets = 1
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=7, last_window=last_window)

    expected = pd.Series(
        data  = np.array([0.73799541, 0.18249173, 0.17545176, 0.53155137, 0.53182759,
                          0.73799541, 0.18249173]),
        index = pd.date_range(start='2000-08-15', periods=7, freq='D'),
        name  = 'pred'
    )

    pd.testing.assert_series_equal(predictions, expected)


def test_predict_7_steps_with_offset_Day_5_n_offsets_2_using_last_window():
    """
    Test predict method with offset pd.offsets.Day(5) and n_offsets=2,
    using last_window.
    """
    y_monthly = y.copy()
    y_monthly.index = pd.date_range(start='1990-01-01', periods=50, freq='MS')
    last_window = pd.Series(
        data = np.array([0.73799541, 0.18249173, 0.17545176, 0.53155137, 0.53182759,
                         0.25045537, 0.48303426, 0.98555979, 0.51948512, 0.61289453]),
        index = pd.date_range(start='2000-08-05', periods=10, freq='MS'),
        name = 'y'
    )

    forecaster = ForecasterEquivalentDate(
                     offset    = pd.DateOffset(months=2),
                     n_offsets = 2,
                     agg_func  = np.mean
                 )
    forecaster.fit(y=y_monthly)
    predictions = forecaster.predict(steps=7, last_window=last_window)

    expected = pd.Series(
        data  = np.array([0.50125969, 0.79922716, 0.50125969, 0.79922716,
                          0.50125969, 0.79922716, 0.50125969]),
        index = pd.date_range(start='2001-07-01', periods=7, freq='MS'),
        name  = 'pred'
    )

    pd.testing.assert_series_equal(predictions, expected)


# Unit test set_params ForecasterDirect
# ==============================================================================
from skforecast.direct import ForecasterDirect
from sklearn.linear_model import LinearRegression


def test_set_params():
    """
    """
    forecaster = ForecasterDirect(LinearRegression(fit_intercept=True), lags=3, steps=3)
    new_params = {'fit_intercept': False}
    forecaster.set_params(new_params)
    expected = {'copy_X': True,
                 'fit_intercept': False,
                 'n_jobs': None,
                 'normalize': 'deprecated',  # For sklearn < 1.2
                 'positive': False
                }
    results = forecaster.regressor.get_params()
    results.update({'normalize': 'deprecated'})
    
    assert results == expected


# Unit test set_lags ForecasterDirectMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirectMultiVariate


@pytest.mark.parametrize("lags", 
                         [None, {'l1': None, 'l2': None}], 
                         ids = lambda lags: f'lags: {lags}')
def test_set_lags_ValueError_when_lags_set_to_None_and_window_features_is_None(lags):
    """
    Test ValueError is raised when lags is set to None and window_features is None.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor       = LinearRegression(),
                     level           = 'l1',
                     steps           = 3,
                     lags            = 3,
                     window_features = None
                 )
    err_msg = re.escape(
        "At least one of the arguments `lags` or `window_features` "
        "must be different from None. This is required to create the "
        "predictors used in training the forecaster."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_lags(lags=lags)


@pytest.mark.parametrize("lags", 
                         [3, [1, 2, 3], np.array([1, 2, 3]), range(1, 4)],
                         ids = lambda lags: f'lags: {lags}')
def test_set_lags_with_different_inputs(lags):
    """
    Test how lags and max_lag attributes change with lags argument of different types.
    """
    forecaster = ForecasterDirectMultiVariate(
        LinearRegression(), level='l1', lags=3, steps=2
    )
    forecaster.set_lags(lags=lags)

    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3]))
    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3']
    assert forecaster.max_lag == 3
    assert forecaster.window_size == 3


def test_set_lags_when_lags_argument_is_a_dict():
    """
    Test how lags and max_lag attributes change when lags argument is a dict.
    """
    forecaster = ForecasterDirectMultiVariate(LinearRegression(), level='l1', 
                                               lags=3, steps=2)
    forecaster.set_lags(lags={'l1': 3, 'l2': [1, 5]})
    
    np.testing.assert_array_almost_equal(forecaster.lags['l1'], np.array([1, 2, 3]))
    np.testing.assert_array_almost_equal(forecaster.lags['l2'], np.array([1, 5]))
    assert forecaster.lags_names['l1'] == ['l1_lag_1', 'l1_lag_2', 'l1_lag_3']
    assert forecaster.lags_names['l2'] == ['l2_lag_1', 'l2_lag_5']
    assert forecaster.max_lag == 5
    assert forecaster.window_size == 5


@pytest.mark.parametrize("lags", 
                         [{'l1': 3, 'l2': None}, {'l1': 3, 'l2': []}],
                         ids = lambda lags: f'lags: {lags}')
def test_set_lags_when_lags_argument_is_a_dict_with_None(lags):
    """
    Test how lags and max_lag attributes change when lags argument is a dict 
    containing None.
    """
    forecaster = ForecasterDirectMultiVariate(LinearRegression(), level='l1', 
                                               lags=3, steps=2)
    forecaster.set_lags(lags=lags)
    
    np.testing.assert_array_almost_equal(forecaster.lags['l1'], np.array([1, 2, 3]))
    assert forecaster.lags['l2'] is None
    assert forecaster.lags_names['l1'] == ['l1_lag_1', 'l1_lag_2', 'l1_lag_3']
    assert forecaster.lags_names['l2'] is None
    assert forecaster.max_lag == 3
    assert forecaster.window_size == 3


@pytest.mark.parametrize("lags, max_lag",
                         [(3, 3), 
                          ({'l1': 3, 'l2': 4}, 4), 
                          ({'l1': None, 'l2': 5}, 5), 
                          ({'l1': [], 'l2': 5}, 5)],
                         ids = lambda value: f'lags: {value}')
def test_set_lags_max_lag_stored(lags, max_lag):
    """
    Test max_lag is equal to the maximum lag.
    """
    forecaster = ForecasterDirectMultiVariate(
        LinearRegression(), level='l1', lags=1, steps=2
    )
    forecaster.set_lags(lags=lags)
    
    assert forecaster.max_lag == max_lag
    assert forecaster.window_size == max_lag


def test_set_lags_when_differentiation_is_not_None():
    """
    Test how `window_size` is also updated when the forecaster includes 
    differentiation.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor       = LinearRegression(),
                     level           = 'l1',
                     lags            = 3,
                     steps           = 2,
                     differentiation = 1
                 )
    
    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3]))
    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3']
    assert forecaster.max_lag == 3
    assert forecaster.window_size == 3 + 1
    assert forecaster.differentiator.window_size == 3 + 1
    
    forecaster.set_lags(lags=5)

    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3, 4, 5]))
    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
    assert forecaster.max_lag == 5
    assert forecaster.window_size == 5 + 1
    assert forecaster.differentiator.window_size == 5 + 1


def test_set_lags_when_window_features():
    """
    Test how `window_size` is also updated when the forecaster includes
    window_features.
    """
    rolling = RollingFeatures(stats='mean', window_sizes=6)
    forecaster = ForecasterDirectMultiVariate(
        LinearRegression(), level='l1', steps=2, lags=9, window_features=rolling
    )
    
    forecaster.set_lags(lags=5)

    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3, 4, 5]))
    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
    assert forecaster.max_lag == 5
    assert forecaster.max_size_window_features == 6
    assert forecaster.window_size == 6


@pytest.mark.parametrize("lags", 
                         [None, []],
                         ids = lambda lags: f'lags: {lags}')
def test_set_lags_to_None(lags):
    """
    Test how lags and max_lag attributes change when lags is set to None.
    """
    rolling = RollingFeatures(stats='mean', window_sizes=3)
    forecaster = ForecasterDirectMultiVariate(
        LinearRegression(), level='l1', steps=3, lags=5, window_features=rolling
    )
    
    forecaster.set_lags(lags=lags)

    assert forecaster.lags is None
    assert forecaster.lags_names is None
    assert forecaster.max_lag is None
    assert forecaster.max_size_window_features == 3
    assert forecaster.window_size == 3


@pytest.mark.parametrize("lags", 
                         [{'l1': None, 'l2': None}, {'l1': [], 'l2': []}],
                         ids = lambda lags: f'lags: {lags}')
def test_set_lags_to_None_dict(lags):
    """
    Test how lags and max_lag attributes change when lags is set to None with a dict.
    """
    rolling = RollingFeatures(stats='mean', window_sizes=3)
    forecaster = ForecasterDirectMultiVariate(
        LinearRegression(), level='l1', steps=3, lags=5, window_features=rolling
    )
    
    forecaster.set_lags(lags=lags)

    assert forecaster.lags == {'l1': None, 'l2': None}
    assert forecaster.lags_names == {'l1': None, 'l2': None}
    assert forecaster.max_lag is None
    assert forecaster.max_size_window_features == 3
    assert forecaster.window_size == 3



# Unit test predict interval method
# ==============================================================================
import os
import keras
import numpy as np
import pandas as pd

from skforecast.deep_learning import ForecasterRnn
from skforecast.deep_learning.utils import create_and_compile_model

os.environ["KERAS_BACKEND"] = "torch"

series = pd.DataFrame(
    {
        "1": pd.Series(np.arange(50)),
        "2": pd.Series(np.arange(50)),
        "3": pd.Series(np.arange(50)),
    }
)
lags = 3
steps = 4
levels = ["1", "2"]
activation = "relu"
optimizer = keras.optimizers.Adam(learning_rate=0.01)
loss = keras.losses.MeanSquaredError()
recurrent_units = 100
dense_units = [128, 64]


model = create_and_compile_model(
    series=series,
    lags=lags,
    steps=steps,
    levels=levels,
    recurrent_units=recurrent_units,
    dense_units=dense_units,
    activation=activation,
    optimizer=optimizer,
    loss=loss,
)

# Adjust after including steps by default
"""
def test_predict_interval_output_size_with_steps_by_default():
    
    #Test output sizes for predicting steps defined by default with intervals
    
    # Create a ForecasterRnn object
    forecaster = ForecasterRnn(model, levels, lags=lags)
    forecaster.fit(series)

    # Call the predict method
    int_preds = forecaster.predict_interval()

    # Check the shape and values of the predictions
    assert int_preds.shape == (steps * len(levels), 4)
    """


def test_predict_interval_output_size_3_steps_ahead():
    """
    Test output sizes for predicting 3 steps ahead with intervals
    """
    # Create a ForecasterRnn object
    forecaster = ForecasterRnn(model, levels, lags=lags)
    forecaster.fit(series)

    # Call the predict method
    int_preds = forecaster.predict_interval(steps=3)

    # Check the shape and values of the predictions
    assert int_preds.shape == (3 * len(levels), 4)


def test_predict_interval_output_size_2_steps_ahead_specific_levels():
    """
    Test output sizes for predicting 2 steps ahead with intervals and specific levels
    """
    # Create a ForecasterRnn object
    forecaster = ForecasterRnn(model, levels, lags=lags)
    forecaster.fit(series)

    # Call the predict method
    int_preds = forecaster.predict_interval(steps=2, levels="1")

    # Check the shape and values of the predictions
    assert int_preds.shape == (2 * 1, 4)




from .preprocessing import (
    TimeSeriesDifferentiator,
    DateTimeFeatureTransformer,
    create_datetime_features,
    exog_long_to_dict,
    series_long_to_dict,
    RollingFeatures,
    QuantileBinner,
    FastOrdinalEncoder,
    ConformalIntervalCalibrator,
)



################################################################################
#                      skforecast.ForecasterRnn.utils                          #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License      #
################################################################################
# coding=utf-8

from __future__ import annotations
from typing import Union, Optional
import warnings
import pandas as pd 
from ..utils import check_optional_dependency

try:
    import keras
    from keras.models import Model
    from keras.layers import Dense, Input, LSTM, SimpleRNN
    from keras.optimizers import Adam
    from keras.losses import MeanSquaredError
except Exception as e:
    package_name = str(e).split(" ")[-1].replace("'", "")
    check_optional_dependency(package_name=package_name)


def create_and_compile_model(
    series: pd.DataFrame,
    lags: Union[int, list],
    steps: Union[int, list],
    levels: Optional[Union[str, int, list]] = None,
    recurrent_layer: str = "LSTM",
    recurrent_units: Union[int, list] = 100,
    dense_units: Union[int, list] = 64,
    activation: Union[str, dict] = "relu",
    optimizer: object = Adam(learning_rate=0.01),
    loss: object = MeanSquaredError(),
    compile_kwargs: dict = {},
) -> keras.models.Model:
    """
    Creates a neural network model for time series prediction with flexible recurrent layers.

    Parameters
    ----------
    series : pandas DataFrame
        Input time series.
    lags : int, list
        Number of lagged time steps to consider in the input, or a list of 
        specific lag indices.
    steps : int, list
        Number of steps to predict into the future, or a list of specific step 
        indices.
    levels : str, int, list, default `None`
        Number of output levels (features) to predict, or a list of specific 
        level indices. If None, defaults to the number of input series.
    recurrent_layer : str, default `'LSTM'`
        Type of recurrent layer to be used ('LSTM' or 'RNN').
    recurrent_units : int, list, default `100`
        Number of units in the recurrent layer(s). Can be an integer or a 
        list of integers for multiple layers.
    dense_units : int, list, default `64`
        List of integers representing the number of units in each dense layer.
    activation : str, dict, default `'relu'`
        Activation function for the recurrent and dense layers. Can be a single 
        string for all layers or a dictionary specifying different activations 
        for 'recurrent_units' and 'dense_units'.
    optimizer : object, default `Adam(learning_rate=0.01)`
        Optimization algorithm and learning rate.
    loss : object, default `MeanSquaredError()`
        Loss function for model training.
    compile_kwargs : dict, default `{}` 
        Additional arguments for model compilation.

    Returns
    -------
    model : keras.models.Model
        Compiled neural network model.
    
    Raises
    ------
    TypeError
        If any of the input arguments are of incorrect type.
    ValueError
        If the activation dictionary does not have the required keys or if the 
        lengths of the lists in the activation dictionary do not match the 
        corresponding parameters.
    
    """

    if keras.__version__ > "3":
        print(f"keras version: {keras.__version__}")
        print(f"Using backend: {keras.backend.backend()}")
        if keras.backend.backend() == "tensorflow":
            import tensorflow
            print(f"tensorflow version: {tensorflow.__version__}")
        elif keras.backend.backend() == "torch":
            import torch
            print(f"torch version: {torch.__version__}")
        elif keras.backend.backend() == "jax":
            import jax
            print(f"jax version: {jax.__version__}")
        else:
            print("Backend not recognized")

    err_msg = f"`series` must be a pandas DataFrame. Got {type(series)}."

    if not isinstance(series, pd.DataFrame):
        raise TypeError(err_msg)

    n_series = series.shape[1]

    # Dense units must be a list, None or int
    if not isinstance(dense_units, (list, int, type(None))):
        raise TypeError(
            f"`dense_units` argument must be a list or int. Got {type(dense_units)}."
        )
    if isinstance(dense_units, int):
        dense_units = [dense_units]

    # Recurrent units must be a list or int
    if not isinstance(recurrent_units, (list, int)):
        raise TypeError(
            f"`recurrent_units` argument must be a list or int. Got {type(recurrent_units)}."
        )
    if isinstance(recurrent_units, int):
        recurrent_units = [recurrent_units]

    # Lags, steps and levels must be int or list
    if not isinstance(lags, (int, list)):
        raise TypeError(f"`lags` argument must be a list or int. Got {type(lags)}.")
    if not isinstance(steps, (int, list)):
        raise TypeError(f"`steps` argument must be a list or int. Got {type(steps)}.")
    if not isinstance(levels, (str, int, list, type(None))):
        raise TypeError(
            f"`levels` argument must be a string, list or int. Got {type(levels)}."
        )

    if isinstance(lags, list):
        lags = len(lags)
    if isinstance(steps, list):
        steps = len(steps)
    if isinstance(levels, list):
        levels = len(levels)
    elif isinstance(levels, (str)):
        levels = 1
    elif isinstance(levels, type(None)):
        levels = series.shape[1]
    elif isinstance(levels, int):
        pass
    else:
        raise TypeError(
            f"`levels` argument must be a string, list or int. Got {type(levels)}."
        )

    if isinstance(activation, str):
        if dense_units is not None:
            activation = {
                "recurrent_units": [activation] * len(recurrent_units),
                "dense_units": [activation] * len(dense_units),
            }
        else:
            activation = {
                "recurrent_units": [activation] * len(recurrent_units)
            }
    elif isinstance(activation, dict):
        # Check if the dictionary has the required keys
        if "recurrent_units" not in activation.keys():
            raise ValueError("The activation dictionary must have a 'recurrent_units' key.")
        if dense_units is not None and "dense_units" not in activation.keys():
            raise ValueError("The activation dictionary must have a 'dense_units' key if dense_units is not None.")
        # Check if the values are lists
        if not isinstance(activation["recurrent_units"], list):
            raise TypeError("The 'recurrent_units' value in the activation dictionary must be a list.")
        if dense_units is not None and not isinstance(activation["dense_units"], list):
            raise TypeError("The 'dense_units' value in the activation dictionary must be a list if dense_units is not None.")
        # Check if the lists have the same length as the corresponding parameters
        if len(activation["recurrent_units"]) != len(recurrent_units):
            raise ValueError("The 'recurrent_units' list in the activation dictionary must have the same length as the recurrent_units parameter.")
        if dense_units is not None and len(activation["dense_units"]) != len(dense_units):
            raise ValueError("The 'dense_units' list in the activation dictionary must have the same length as the dense_units parameter.")
    else:
        raise TypeError(f"`activation` argument must be a string or dict. Got {type(activation)}.")

    input_layer = Input(shape=(lags, n_series))
    x = input_layer

    # Dynamically create multiple recurrent layers if recurrent_units is a list
    if isinstance(recurrent_units, list):
        for i, units in enumerate(recurrent_units[:-1]):  # All layers except the last one
            if recurrent_layer == "LSTM":
                x = LSTM(units, activation=activation["recurrent_units"][i], return_sequences=True)(x)
            elif recurrent_layer == "RNN":
                x = SimpleRNN(units, activation=activation["recurrent_units"][i], return_sequences=True)(x)
            else:
                raise ValueError(f"Invalid recurrent layer: {recurrent_layer}")
        # Last layer without return_sequences
        if recurrent_layer == "LSTM":
            x = LSTM(recurrent_units[-1], activation=activation["recurrent_units"][-1])(x)
        elif recurrent_layer == "RNN":
            x = SimpleRNN(recurrent_units[-1], activation=activation["recurrent_units"][-1])(x)
        else:
            raise ValueError(f"Invalid recurrent layer: {recurrent_layer}")
    else:
        # Single recurrent layer
        if recurrent_layer == "LSTM":
            x = LSTM(recurrent_units, activation=activation["recurrent_units"][0])(x)
        elif recurrent_layer == "RNN":
            x = SimpleRNN(recurrent_units, activation=activation["recurrent_units"][0])(x)
        else:
            raise ValueError(f"Invalid recurrent layer: {recurrent_layer}")

    # Dense layers
    if dense_units is not None:
        for i, nn in enumerate(dense_units):
            x = Dense(nn, activation=activation["dense_units"][i])(x)

    # Output layer
    x = Dense(levels * steps, activation="linear")(x)
    # model = Model(inputs=input_layer, outputs=x)
    output_layer = keras.layers.Reshape((steps, levels))(x)
    model = Model(inputs=input_layer, outputs=output_layer)

    # Compile the model if optimizer, loss or compile_kwargs are passed
    if optimizer is not None or loss is not None or compile_kwargs:
        # give more priority to the parameters passed in the function check if the
        # parameters passes in compile_kwargs include optimizer and loss if so,
        # delete them from compile_kwargs and raise a warning
        if "optimizer" in compile_kwargs.keys():
            compile_kwargs.pop("optimizer")
            warnings.warn("`optimizer` passed in `compile_kwargs`. Ignoring it.")
        if "loss" in compile_kwargs.keys():
            compile_kwargs.pop("loss")
            warnings.warn("`loss` passed in `compile_kwargs`. Ignoring it.")

        model.compile(optimizer=optimizer, loss=loss, **compile_kwargs)

    return model



# Unit test preprocess_last_window
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.exceptions import IndexWarning
from skforecast.utils import preprocess_last_window


def test_output_preprocess_last_window_when_last_window_index_is_DatetimeIndex_and_has_frequency():
    """
    Test values returned by when last_window is a pandas Series with DatetimeIndex
    and freq is not None.
    """
    last_window = pd.Series(
            data = np.arange(3),
            index = pd.date_range("1990-01-01", periods=3, freq='D')
        )
    results = preprocess_last_window(last_window)
    expected = (np.arange(3),
                pd.DatetimeIndex(['1990-01-01', '1990-01-02', '1990-01-03'],
                                 dtype='datetime64[ns]', freq='D')
               )
    
    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()


def test_output_preprocess_last_window_when_last_window_index_is_RangeIndex():
    """
    Test values returned by when last_window is a pandas Series with RangeIndex
    """
    last_window = pd.Series(
            data = np.arange(3),
            index = pd.RangeIndex(start=0, stop=3, step=1)
        )
    results = preprocess_last_window(last_window)
    expected = (np.arange(3),
                pd.RangeIndex(start=0, stop=3, step=1)
               )
    
    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()


def test_output_preprocess_last_window_when_last_window_index_is_DatetimeIndex_but_has_not_frequency():
    """
    Test values returned by when last_window is a pandas Series with DatetimeIndex
    and freq is None.
    """
    last_window = pd.Series(
            data = np.arange(3),
            index = pd.to_datetime(["1990-01-01", "1990-01-02", "1990-01-03"])
        )
    warn_msg = re.escape(
        "`last_window` has a pandas DatetimeIndex without a frequency. The index "
        "will be replaced by a RangeIndex starting from 0 with a step of 1. "
        "To avoid this warning, set the frequency of the DatetimeIndex using "
        "`last_window = last_window.asfreq('desired_frequency', fill_value=np.nan)`."
    )
    with pytest.warns(IndexWarning, match=warn_msg):
        results = preprocess_last_window(last_window)

    expected = (np.arange(3),
                pd.RangeIndex(start=0, stop=3, step=1)
               )
    
    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()
    
    
def test_output_preprocess_last_window_when_last_window_index_is_not_DatetimeIndex_or_RangeIndex():
    """
    Test values returned by when last_window is a pandas Series without DatetimeIndex or RangeIndex.
    """
    last_window = pd.Series(data=np.arange(3), index=['0', '1', '2'])
    warn_msg = re.escape(
        "`last_window` has an unsupported index type (not pandas DatetimeIndex or "
        "RangeIndex). The index will be replaced by a RangeIndex starting "
        "from 0 with a step of 1. To avoid this warning, ensure that "
        "`last_window.index` is a DatetimeIndex with a frequency or a RangeIndex."
    )
    with pytest.warns(IndexWarning, match=warn_msg):
        results = preprocess_last_window(last_window)

    expected = (np.arange(3),
                pd.RangeIndex(start=0, stop=3, step=1)
               )
    
    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()



# Unit test backtesting ForecasterEquivalentDate
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
from skforecast.recursive import ForecasterEquivalentDate
from skforecast.model_selection._validation import backtesting_forecaster
from skforecast.model_selection._split import TimeSeriesFold

# Fixtures
from .fixtures_forecaster_equivalent_date import y


@pytest.mark.parametrize("initial_train_size", 
                         [len(y) - 20, "2000-01-30 00:00:00"],
                         ids=lambda init: f'initial_train_size: {init}')
def test_backtesting_with_ForecasterEquivalentDate(initial_train_size):
    """
    Test backtesting with ForecasterEquivalentDate.
    """
    forecaster = ForecasterEquivalentDate(
                     offset    = pd.DateOffset(days=10),
                     n_offsets = 2 
                 )
    cv = TimeSeriesFold(
             initial_train_size = initial_train_size,
             steps              = 5,
             refit              = True,
         )

    metric, predictions = backtesting_forecaster(
        forecaster = forecaster,
        y          = y,
        cv         = cv,
        metric     = 'mean_absolute_error',
        verbose    = False,
        n_jobs     = 'auto'
    )
    expected_metric = pd.DataFrame({'mean_absolute_error': [0.2537094475]})
    expected = pd.DataFrame(
        data    = np.array([0.48878949, 0.78924075, 0.58151378, 0.3353507, 0.56024382,
                            0.53047716, 0.27214019, 0.20185749, 0.41263271, 0.58140185,
                            0.36325295, 0.64156648, 0.57765904, 0.5523543, 0.57413684,
                            0.31761006, 0.39406999, 0.56082619, 0.61893703, 0.5664064]),
        index   = pd.date_range(start='2000-01-31', periods=20, freq='D'),
        columns = ['pred']
    )

    pd.testing.assert_frame_equal(metric, expected_metric)
    pd.testing.assert_frame_equal(predictions, expected)



# Unit test _create_predict_inputs ForecasterDirect
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.compose import make_column_transformer
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirect

# Fixtures
from .fixtures_forecaster_direct import y as y_categorical
from .fixtures_forecaster_direct import exog as exog_categorical
from .fixtures_forecaster_direct import data  # to test results when using differentiation


@pytest.mark.parametrize("steps", [[1, 2.0, 3], [1, 4.]], 
                         ids=lambda steps: f'steps: {steps}')
def test_create_predict_inputs_TypeError_when_steps_list_contain_floats(steps):
    """
    Test _create_predict_inputs TypeError when steps is a list with floats.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)
    forecaster.fit(y=pd.Series(np.arange(10)))

    err_msg = re.escape(
        (f"`steps` argument must be an int, a list of ints or `None`. "
         f"Got {type(steps)}.")
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster._create_predict_inputs(steps=steps)


def test_create_predict_inputs_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)

    err_msg = re.escape(
        ("This Forecaster instance is not fitted yet. Call `fit` with "
         "appropriate arguments before using predict.")
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster._create_predict_inputs(steps=5)


@pytest.mark.parametrize("steps", [3, [1, 2, 3], None], 
                         ids=lambda steps: f'steps: {steps}')
def test_create_predict_inputs_output(steps):
    """
    Test _create_predict_inputs output.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=3)
    forecaster.fit(y=pd.Series(np.arange(50, dtype=float), 
                   index=pd.date_range(start='2020-01-01', periods=50, freq='D')))
    results = forecaster._create_predict_inputs(steps=steps)

    expected = (
        [np.array([[49., 48., 47.]]),
         np.array([[49., 48., 47.]]),
         np.array([[49., 48., 47.]])],
        ['lag_1', 'lag_2', 'lag_3'],
        [1, 2, 3],
        pd.date_range(start='2020-02-20', periods=3, freq='D')
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_when_regressor_is_LinearRegression_with_list_interspersed():
    """
    Test _create_predict_inputs output when steps is
    a list with interspersed steps.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)
    forecaster.fit(y=pd.Series(np.arange(50, dtype=float)))
    results = forecaster._create_predict_inputs(steps=[1, 4])

    expected = (
        [np.array([[49., 48., 47.]]),
         np.array([[49., 48., 47.]])],
        ['lag_1', 'lag_2', 'lag_3'],
        [1, 4],
        pd.RangeIndex(start=50, stop=54, step=1)[[0, 3]]
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_when_last_window():
    """
    Test _create_predict_inputs output when external last_window.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)
    forecaster.fit(y=pd.Series(np.arange(50, dtype=float)))
    last_window = pd.Series(
        data  = [47, 48, 49], 
        index = pd.RangeIndex(start=47, stop=50, step=1)
    )
    results = forecaster._create_predict_inputs(steps=[1, 2, 3, 4], last_window=last_window)

    expected = (
        [np.array([[49., 48., 47.]]),
         np.array([[49., 48., 47.]]),
         np.array([[49., 48., 47.]]),
         np.array([[49., 48., 47.]])],
        ['lag_1', 'lag_2', 'lag_3'],
        [1, 2, 3, 4],
        pd.RangeIndex(start=50, stop=54, step=1)
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_when_exog():
    """
    Test _create_predict_inputs output when exog.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)
    forecaster.fit(
        y=pd.Series(np.arange(50, dtype=float)),
        exog=pd.Series(np.arange(start=100, stop=150, step=1), name="exog")
    )
    results = forecaster._create_predict_inputs(
                  steps = 5, 
                  exog  = pd.Series(np.arange(start=25, stop=50, step=0.5, dtype=float),
                                    index=pd.RangeIndex(start=50, stop=100),
                                    name="exog")
               )

    expected = (
        [np.array([[49., 48., 47., 25.]]),
         np.array([[49., 48., 47., 25.5]]),
         np.array([[49., 48., 47., 26.]]),
         np.array([[49., 48., 47., 26.5]]),
         np.array([[49., 48., 47., 27.]])],
        ['lag_1', 'lag_2', 'lag_3', 'exog'],
        [1, 2, 3, 4, 5],
        pd.RangeIndex(start=50, stop=55, step=1)
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_with_transform_y():
    """
    Test _create_predict_inputs output when StandardScaler.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    transformer_y = StandardScaler()

    forecaster = ForecasterDirect(
                     regressor     = LinearRegression(),
                     lags          = 5,
                     steps         = 5,
                     transformer_y = transformer_y,
                 )
    forecaster.fit(y=y)
    results = forecaster._create_predict_inputs()

    expected = (
        [np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608]]),
         np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608]]),
         np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608]]),
         np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608]]),
         np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608]])],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5'],
        [1, 2, 3, 4, 5],
        pd.RangeIndex(start=20, stop=25, step=1)
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'], 
                         ids=lambda n_jobs: f'n_jobs: {n_jobs}')
def test_create_predict_inputs_output_with_transform_y_and_transform_exog(n_jobs):
    """
    Test _create_predict_inputs output when StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    exog = pd.DataFrame({
                'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4, 60.3, 87.2,
                          7.5, 60.4, 50.3, 57.3, 24.7, 87.4, 87.2, 60.4, 50.7, 7.5],
                'col_2': ['a'] * 10 + ['b'] * 10}
           )
    exog_predict = exog.copy()
    exog_predict.index = pd.RangeIndex(start=20, stop=40)

    transformer_y = StandardScaler()
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['col_1']),
                            ('onehot', OneHotEncoder(), ['col_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    
    forecaster = ForecasterDirect(
                     regressor        = LinearRegression(),
                     lags             = 5,
                     steps            = 5,
                     transformer_y    = transformer_y,
                     transformer_exog = transformer_exog,
                     n_jobs           = n_jobs
                 )
    forecaster.fit(y=y, exog=exog)
    results = forecaster._create_predict_inputs(steps=[1, 2, 3, 4, 5], exog=exog_predict)

    expected = (
        [np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608,
                    -1.7093071, 1., 0.,]]),
         np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608,
                    -1.0430105, 1., 0.,]]),
         np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608,
                    0.372377, 1., 0.,]]),
         np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608,
                    0.2540995, 1., 0.,]]),
         np.array([[-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608,
                    -0.006111, 1., 0.,]])],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'col_1', 'col_2_a', 'col_2_b'],
        [1, 2, 3, 4, 5],
        pd.RangeIndex(start=20, stop=25, step=1)
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_when_categorical_features_native_implementation_HistGradientBoostingRegressor():
    """
    Test _create_predict_inputs output when using HistGradientBoostingRegressor 
    and categorical variables.
    """
    df_exog = pd.DataFrame({
        'exog_1': exog_categorical,
        'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
        'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)}
    )
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterDirect(
                     regressor        = HistGradientBoostingRegressor(
                                            categorical_features = categorical_features,
                                            random_state         = 123
                                        ),
                     lags             = 5,
                     steps            = 10, 
                     transformer_y    = None,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_categorical, exog=df_exog)
    results = forecaster._create_predict_inputs(steps=10, exog=exog_predict)

    expected = (
        [np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    0., 0., 0.12062867]]),
         np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    1., 1., 0.8263408 ]]),
         np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    2., 2., 0.60306013]]),
         np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    3., 3., 0.54506801]]),
         np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    4., 4., 0.34276383]]),
         np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    0., 0., 0.30412079]]),
         np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    1., 1., 0.41702221]]),
         np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    2., 2., 0.68130077]]),
         np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    3., 3., 0.87545684]]),
         np.array([[0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    4., 4., 0.51042234]])],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_2', 'exog_3', 'exog_1'],
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        pd.RangeIndex(start=50, stop=60, step=1)
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_when_regressor_is_LinearRegression_with_exog_differentiation_is_1():
    """
    Test _create_predict_inputs when using LinearRegression as regressor 
    and differentiation=1.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )

    forecaster = ForecasterDirect(
                     regressor       = LinearRegression(),
                     steps           = 5,
                     lags            = 15,
                     differentiation = 1
                )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    results = forecaster._create_predict_inputs(exog=exog.loc[end_train:])
    
    expected = (
        [np.array([[ 0.07503713, -0.48984868, -0.01463081,  0.10597971, -0.01018012,
                0.02377842,  0.09883014,  0.01630394,  0.17596768, -0.00584249,
                0.09807633,  0.04869748,  0.07558021, -0.56028323,  0.14355438,
                1.16172882]]),
        np.array([[ 0.07503713, -0.48984868, -0.01463081,  0.10597971, -0.01018012,
                0.02377842,  0.09883014,  0.01630394,  0.17596768, -0.00584249,
                0.09807633,  0.04869748,  0.07558021, -0.56028323,  0.14355438,
                0.29468848]]),
        np.array([[ 0.07503713, -0.48984868, -0.01463081,  0.10597971, -0.01018012,
                0.02377842,  0.09883014,  0.01630394,  0.17596768, -0.00584249,
                0.09807633,  0.04869748,  0.07558021, -0.56028323,  0.14355438,
                -0.4399757 ]]),
        np.array([[ 0.07503713, -0.48984868, -0.01463081,  0.10597971, -0.01018012,
                0.02377842,  0.09883014,  0.01630394,  0.17596768, -0.00584249,
                0.09807633,  0.04869748,  0.07558021, -0.56028323,  0.14355438,
                1.25008389]]),
        np.array([[ 0.07503713, -0.48984868, -0.01463081,  0.10597971, -0.01018012,
                0.02377842,  0.09883014,  0.01630394,  0.17596768, -0.00584249,
                0.09807633,  0.04869748,  0.07558021, -0.56028323,  0.14355438,
                1.37496887]])],
        ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'lag_6', 'lag_7', 'lag_8',
         'lag_9', 'lag_10', 'lag_11', 'lag_12', 'lag_13', 'lag_14', 'lag_15', 'exog'],
        [1, 2, 3, 4, 5],
        pd.date_range(start='2003-04-01', periods=5, freq='MS')
    )
    
    for step in range(len(results[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_window_features():
    """
    Test _create_predict_inputs output with window_features.
    """
    y_datetime = pd.Series(
        np.arange(50), index=pd.date_range('2020-01-01', periods=50, freq='D'),
        name='y', dtype=float
    )
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[4, 5, 6]
    )
    forecaster = ForecasterDirect(
        LinearRegression(), steps=3, lags=3, window_features=rolling
    )
    forecaster.fit(y=y_datetime)
    results = forecaster._create_predict_inputs()

    expected = (
        [np.array([[49., 48., 47., 47.5, 47., 279.]]),
         np.array([[49., 48., 47., 47.5, 47., 279.]]),
         np.array([[49., 48., 47., 47.5, 47., 279.]])],
        ['lag_1', 'lag_2', 'lag_3', 'roll_mean_4', 'roll_median_5', 'roll_sum_6'],
        [1, 2, 3],
        pd.date_range(start='2020-02-20', periods=3, freq='D')
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_with_2_window_features():
    """
    Test _create_predict_inputs output with 2 window_features.
    """
    y_datetime = pd.Series(
        np.arange(50), index=pd.date_range('2020-01-01', periods=50, freq='D'),
        name='y', dtype=float
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[4, 5])
    rolling_2 = RollingFeatures(stats=['sum'], window_sizes=6)
    forecaster = ForecasterDirect(
        LinearRegression(), steps=3, lags=3, window_features=[rolling, rolling_2]
    )
    forecaster.fit(y=y_datetime)
    results = forecaster._create_predict_inputs()

    expected = (
        [np.array([[49., 48., 47., 47.5, 47., 279.]]),
         np.array([[49., 48., 47., 47.5, 47., 279.]]),
         np.array([[49., 48., 47., 47.5, 47., 279.]])],
        ['lag_1', 'lag_2', 'lag_3', 'roll_mean_4', 'roll_median_5', 'roll_sum_6'],
        [1, 2, 3],
        pd.date_range(start='2020-02-20', periods=3, freq='D')
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])


def test_create_predict_inputs_output_window_features_and_no_lags():
    """
    Test _create_predict_inputs output with window_features and no lags.
    """
    y_datetime = pd.Series(
        np.arange(50), index=pd.date_range('2020-01-01', periods=50, freq='D'),
        name='y', dtype=float
    )
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[4, 5, 6]
    )
    forecaster = ForecasterDirect(
        LinearRegression(), steps=3, lags=None, window_features=rolling
    )
    forecaster.fit(y=y_datetime)
    results = forecaster._create_predict_inputs()

    expected = (
        [np.array([[47.5, 47., 279.]]),
         np.array([[47.5, 47., 279.]]),
         np.array([[47.5, 47., 279.]])],
        ['roll_mean_4', 'roll_median_5', 'roll_sum_6'],
        [1, 2, 3],
        pd.date_range(start='2020-02-20', periods=3, freq='D')
    )
    
    for step in range(len(expected[0])):
        np.testing.assert_almost_equal(results[0][step], expected[0][step])
    assert results[1] == expected[1]
    assert results[2] == expected[2]
    pd.testing.assert_index_equal(results[3], expected[3])
